"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2016-2019 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import sys
import time
import struct
import socket
import logging
import calendar
import datetime
import contextlib
import SEAScope.cmds.add_collection
import SEAScope.cmds.add_variable
import SEAScope.cmds.add_granule
import SEAScope.cmds.set_rendering
import SEAScope.cmds.get_rendering
import SEAScope.cmds.set_current_datetime
import SEAScope.cmds.set_altitude
import SEAScope.cmds.look_at
import SEAScope.cmds.zoom_in
import SEAScope.cmds.zoom_out
import SEAScope.cmds.reset_camera
import SEAScope.cmds.select_variable
import SEAScope.cmds.select_variable_by_label
import SEAScope.cmds.get_variable_identifier

logger = logging.getLogger(__name__)


class Throttler(object):
    def __init__(self):
        self.__last_upload = 0
        self.__min_delay_ms = 0

    @property
    def min_delay(self):
        return self.__min_delay_ms

    @min_delay.setter
    def min_delay(self, value):
        logger.debug('Set min delay to {}'.format(value))
        self.__min_delay_ms = max(0, value)

    def apply_delay(self):
        """"""
        now = datetime.datetime.utcnow()
        now_ms = time.mktime(now.timetuple()) * 1e3 + now.microsecond / 1e3
        elapsed = now_ms - self.__last_upload
        if elapsed < self.__min_delay_ms:
            to_wait_ms = self.__min_delay_ms - elapsed
            logger.debug('Delay for {}ms'.format(to_wait_ms))
            time.sleep(to_wait_ms / 1e3)
            now = datetime.datetime.utcnow()
            now_ms = time.mktime(now.timetuple()) * 1e3 + now.microsecond / 1e3
        self.__last_upload = now_ms


throttler = Throttler()


@contextlib.contextmanager
def connect(host, port):
    """Return an open socket which can be used in a "with" statement."""
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        yield s
    except:  # noqa
        _, e, _ = sys.exc_info()
        logger.error(e)
        raise
    finally:
        logger.debug('Closing connection')
        s.send(b'')
        s.shutdown(socket.SHUT_RDWR)
        s.close()


def read_response(link, msg_size):
    """"""
    chunk_size = msg_size
    logger.debug('Message size: {}'.format(msg_size))
    already_read = 0
    result = b''
    while already_read < msg_size:
        buf = link.recv(chunk_size)
        already_read += len(buf)
        result += buf
        chunk_size = msg_size - already_read
    return result


def collection(s, _collection):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.add_collection.serialize
    builder = None
    builder, serialized = serializer(builder, _collection)
    builder.Finish(serialized)
    buf = builder.Output()
    s.sendall(struct.pack('>Q', len(buf)) + buf)


def variable(s, _variable):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.add_variable.serialize
    builder = None
    collection_id = _variable['collection']
    builder, serialized = serializer(builder, collection_id, _variable)
    builder.Finish(serialized)
    buf = builder.Output()
    s.sendall(struct.pack('>Q', len(buf)) + buf)

    if 'rendering' in _variable:
        rendering_config(s, _variable['rendering'])


def rendering_config(s, rcfg):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.set_rendering.serialize
    builder = None
    builder, serialized = serializer(builder, rcfg)
    builder.Finish(serialized)
    buf = builder.Output()
    s.sendall(struct.pack('>Q', len(buf)) + buf)


def granule(s, _granule):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.add_granule.serialize
    builder = None
    builder, serialized = serializer(builder, _granule)
    builder.Finish(serialized)
    buf = builder.Output()
    s.sendall(struct.pack('>Q', len(buf)) + buf)


def current_datetime(link, dt):
    """"""
    global throttler
    throttler.apply_delay()

    timestamp = calendar.timegm(dt.timetuple())
    serializer = SEAScope.cmds.set_current_datetime.serialize
    builder = None
    builder, serialized = serializer(builder, timestamp)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf)) + buf)


def rendering_config_for(link, target):
    """"""
    global throttler
    throttler.apply_delay()

    # Request rendering config from SEAScope
    serializer = SEAScope.cmds.get_rendering.serialize
    deserializer = SEAScope.cmds.get_rendering.deserialize
    builder = None
    builder, serialized = serializer(builder, target)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)

    # Retrieve result
    buf = link.recv(8)
    msg_len = struct.unpack('<Q', buf)
    buf = read_response(link, msg_len[0])
    result = deserializer(buf)
    return result


def altitude(link, altitude):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.set_altitude.serialize
    builder = None
    builder, serialized = serializer(builder, altitude)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def location(link, lon, lat):
    """"""
    global throttler
    throttler.apply_delay()

    lon = ((lon + 180.) % 360.) - 180.
    serializer = SEAScope.cmds.look_at.serialize
    builder = None
    builder, serialized = serializer(builder, lon, lat)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def zoom_in(link):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.zoom_in.serialize
    builder = None
    builder, serialized = serializer(builder)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def zoom_out(link):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.zoom_out.serialize
    builder = None
    builder, serialized = serializer(builder)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def reset_camera(link):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.reset_camera.serialize
    builder = None
    builder, serialized = serializer(builder)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def select_variable_by_id(link, source_id, collection_id, variable_id,
                          selected, exclusive):
    """"""
    global throttler
    throttler.apply_delay()

    _rid = {'granuleLevel': False,
            'sourceId': source_id,
            'collectionId': collection_id,
            'granuleId': 0,
            'variableId': variable_id}

    serializer = SEAScope.cmds.select_variable.serialize
    builder = None
    builder, serialized = serializer(builder, _rid, selected, exclusive)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def select_variable_by_label(link, collection_label, variable_label, selected,
                             exclusive):
    """"""
    global throttler
    throttler.apply_delay()

    serializer = SEAScope.cmds.select_variable_by_label.serialize
    builder = None
    builder, serialized = serializer(builder, collection_label, variable_label,
                                     selected, exclusive)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)


def get_id_for(link, collection_label, variable_label):
    """"""
    global throttler
    throttler.apply_delay()

    # Request rendering config from SEAScope
    serializer = SEAScope.cmds.get_variable_identifier.serialize
    deserializer = SEAScope.cmds.get_variable_identifier.deserialize
    builder = None
    builder, serialized = serializer(builder, collection_label, variable_label)
    builder.Finish(serialized)
    buf = builder.Output()
    link.sendall(struct.pack('>Q', len(buf))+buf)

    # Retrieve result
    buf = link.recv(8)
    msg_len = struct.unpack('<Q', buf)
    buf = read_response(link, msg_len[0])
    result = deserializer(buf)
    return result
