# -*- coding: utf-8 -*-

import os
import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand
try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess


# https://github.com/pydanny/whichcraft/blob/master/whichcraft.py
# BSD-3 clauses licence
def whichcraft(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.
    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.
    Note: This function was backported from the Python 3 source code.
    """
    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def _access_check(fn, mode):
        return (os.path.exists(fn) and os.access(fn, mode) and
                not os.path.isdir(fn))

    # If we're given a path with a directory part, look it up directly
    # rather than referring to PATH directories. This includes checking
    # relative to the current directory, e.g. ./script
    if os.path.dirname(cmd):
        if _access_check(cmd, mode):
            return cmd
        return None

    if path is None:
        path = os.environ.get("PATH", os.defpath)
        if not path:
            return None
        path = path.split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if os.curdir not in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        # See if the given file matches any of the expected path
        # extensions. This will allow us to short circuit when given
        # "python.exe". If it does match, only test that one, otherwise we
        # have to try others.
        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for dir in path:
        normdir = os.path.normcase(dir)
        if normdir not in seen:
            seen.add(normdir)
            for thefile in files:
                name = os.path.join(dir, thefile)
                if _access_check(name, mode):
                    return name
    return None


# Try to retrieve the "which" command implementation
which = None
try:
    from shutil import which
except ImportError:
    try:
        from backports.shutil_which import which
    except ImportError:
        pass

# Fallback to the "whichcraft" implementation
if which is None:
    which = whichcraft


class Tox(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_siote = True

    def run_tests(self):
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)


major_version = '0.2'
package_dir = os.path.dirname(__file__)
version_path = os.path.join(package_dir, 'VERSION.txt')

git = which('git')
if git is not None and os.path.isdir('.git'):
    git_cmd = [git, 'rev-list', 'HEAD', '--count']
    commits = subprocess.check_output(git_cmd).decode('utf-8').strip()
else:
    commits = '0'

version = '{}.{}'.format(major_version, commits)
with open(version_path, 'w') as f:
    f.write('{}\n'.format(version))

readme = ''
with open('README.txt', 'r') as f:
    readme = f.read()

processor_deps = ('matplotlib', 'numpy', 'scipy', 'pyproj')

setup(
    zip_safe=False,
    name='seascope',
    version=version,
    author='Sylvain Herlédan <sylvain.herledan@oceandatalab.com>',
    author_email='seascope@oceandatalab.com',
    maintainer='Sylvain Herlédan <sylvain.herledan@oceandatalab.com>',
    maintainer_email='seascope@oceandatalab.com',
    url='https://seascope.oceandatalab.com/python.html',
    packages=('SEAScope', 'SEAScope.API', 'SEAScope.cmds', 'SEAScope.types',
              'SEAScope.lib', 'SEAScope.cli'),
    scripts=[],
    license='COPYING',
    description='SEAScope Python bindings.',
    long_description=readme,
    install_requires=['flatbuffers'],
    tests_require=['tox', 'virtualenv'],
    extras_require={'processor': processor_deps},
    classifiers=['Development Status :: 4 - Beta',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'Intended Audience :: Education',
                 'Intended Audience :: End Users/Desktop',
                 'Intended Audience :: Science/Research',
                 'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',  # noqa
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: Microsoft :: Windows :: Windows 10',
                 'Operating System :: Microsoft :: Windows :: Windows 8.1',
                 'Operating System :: Microsoft :: Windows :: Windows 8',
                 'Operating System :: Microsoft :: Windows :: Windows 7',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.7',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Scientific/Engineering :: Visualization',
                 'Topic :: Scientific/Engineering :: GIS'],
    cmdclass={'test': Tox},
    entry_points={'console_scripts': (
        'seascope-processor = SEAScope.cli.server:seascope_processor')}
)
